# -*- coding: iso-8859-1 -*-
"""
Diese Modul implementiert ein Hidden-Markow-Model auf der Basis eine
Markow-Modells (Sorry, nur Python 2.3).

Vgl. Skript, "Der Viterbi-Algorithmus"

Anmerkung: Das Programm gibt am Anfang des Programmlaufs jeweils
den verwendeten "Seed" des Zufallszahlengenerators aus.  Wenn ihr
diesen Seed ber die -s-Option des Programms angebt, erhaltet ihr
exakt den gleichen Ablauf (vorausgesetzt, ihr ndert weder Python-Version
noch Hardware noch C-Bibliothek -- unter den Poolmaschinen sollte das
gegeben sein).  Das ist nett, wenn ihr selbst (etwa mit stragischen
prints) nochmal nachsehen wollt, was das HMM getan hat oder wenn ihr
bestimmte Ablufe anderen Leuten zeigen wollt.

Z.B. sollte mit der probD-Funktion als Hauptfunktion, python 2.3,
der glibc 2.2.4 auf einer i386-Architektur folgendes passieren:
examples> python hmm.py -s 1071854081.98 pixelstrip.txt
Random seed is 1071854081.98
( ( ( ( ( ( ( _ _ _ _ _ _ _ _ _ _ _ ) ) ) ) ) ) ) ) ) ) ) ) )
 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 1 0 0 0 0 1
( ( ( ( ( ( _ _ _ _ _ _ _ _ _ _ _ _ ) ) ) ) ) ) ) ) ) ) ) ) )
"""

from __future__ import generators
from markovmodel import Error, MarkovModel, DistFunc, normalize
from markovmodel import ProcessAbsorbed
from sets import Set
import operator, re, random


def interleave(seq1, seq2):
	"""returns an iterator that iterleaves seq2 into seq1, i.e. returns
	the first element of seq1, then the first of seq2, then the second
	of seq1 and so on.
	"""
	curInd = 0
	try:
		while 1:
			yield seq1[curInd]
			yield seq2[curInd]
			curInd += 1
	except IndexError:
		pass


def deInterleave(seq):
	"""does the inverse of interleave, i.e., returns two sequences
	"""
	seq1, seq2 = [], []
	curInd = 0
	try:
		while 1:
			seq1.append(seq[curInd])
			seq2.append(seq[curInd+1])
			curInd += 2
	except IndexError:
		pass
	return seq1, seq2


class HiddenMarkovModel(MarkovModel):
	"""The hidden markov model has an underlying markov model and, for
	each transition, has a distribution of emissions.

	It is constructed with a textual description consisting of three
	sections.  Above a line consisting of three dashes only, there's
	a dictionary as used by MarkovModel.  Between this and the next
	line of three dashes only is a dictionary specifying a distribution
	of the initial state, and finally, for each transition allowed by 
	the model, there is an entry

	+<source-state>-<dest-state>+ <trans-def> +-+ 

	(given our simple parser, +-+ must not be present in trans-def).
	<trans-def> is a dictionary with strings (the emitted symbols,
	henceforth output) as keys and the associated emission probabilities 
	as values.
	"""
	def __init__(self, description):
		markovModel, initDist, rawEmissions = description.split("\n---\n", 2)
		MarkovModel.__init__(self, markovModel)
		self._computeEmissionMatrix(
			self._parseEmissions(rawEmissions))
		self._computeInitDf(initDist)
	
	def _computeInitDf(self, rawInitDist):
		dist = [0]*self.numStates
		d = eval(rawInitDist)
		for stateName in d:
			dist[self._getStateNum(stateName)] = d[stateName]
		self.initDf = DistFunc(dist)

	def _getOutputForNum(self, outputIndex):
		return self.numToOutput[outputIndex]
	
	def _getNumForOutput(self, output):
		return self.outputToNum[output]

	def _outputToOutputNums(self, outputSeq):
		"""returns a sequence of output indices for a sequence of output
		symbols.
		"""
		return map(self._getNumForOutput, outputSeq)

	def _buildEmissionMapping(self, emissionsDict):
		"""builds mappings from emission symbols to their indices and
		vice versa.  Only to be called from _computeEmissionMatrix.
		"""
		allOutputs = {}
		[allOutputs.update(dist) for dist in emissionsDict.itervalues()]
		self.numSymbols = len(allOutputs)
		self.numToOutput = dict(enumerate(allOutputs))
		self.outputToNum = dict(zip(self.numToOutput.values(), 
			self.numToOutput.keys()))

	def _computeEmissionMatrix(self, emissionsDict):
		"""computes a 3d-array (actually, a list of lists of lists)
		for the emission probabilities (src-state, dest-state, output-index).
		"""
		self._buildEmissionMapping(emissionsDict)
		self.emissionMatrix = [[[0]*self.numSymbols
			for i in range(self.numStates)]
				for j in range(self.numStates)]
		for trans, emDist in emissionsDict.iteritems():
			thisLine = self.emissionMatrix[trans[0]][trans[1]]
			for output, prob in emDist.iteritems():
				thisLine[self.outputToNum[output]] = prob
			normalize(thisLine)

	def _parseEmissions(self, rawEmissions):
		"""returns a dictionary of (src,dest)-tuples as keys and
		dictionaries of output -> emission-prob parsed from
		rawEmissions.  This dictionary is then fed to _computeEmissionMatrix.
		"""
		emissions = {}
		for rawEm in map(str.strip, rawEmissions.split("+-+")):
			if not rawEm.strip():
				continue
			srcName, destName, descr = re.match(
				r"(?s)\+([^-]+)-([^+]+)\+(.*)", rawEm).groups()
			emissions[self._getStateNum(srcName), 
				self._getStateNum(destName)] = eval(descr)
		return emissions

	def _computeOutputDfs(self):
		"""computes the dfs for emission outputs.  It creates an attribute
		_outputDfs containing these DFs.  This attribute must be deleted
		when the emission probabilities are changed.
		"""
		self._outputDfs = {}
		noTrans = [0]*self.numSymbols
		for srcInd, srcTrans in enumerate(self.emissionMatrix):
			for destInd, dist in enumerate(srcTrans):
				if dist!=noTrans:
					self._outputDfs[srcInd, destInd
						] = DistFunc(dist)

	def driveOn(self):
		"""drives the HMM one step further, returns an emitted symbol.
		"""
		if not hasattr(self, "_outputDfs"):
			self._computeOutputDfs()
		preState = self.state
		MarkovModel.driveOn(self)
		try:
			return self._getOutputForNum(
				self._outputDfs[(preState, self.state)].draw())
		except KeyError:
			raise Error, "No emissions defined for %s-%s"%tuple(
				map(self._getStateName, [preState, self.state]))
	
	def getAnOutput(self, numSteps):
		"""returns a random output of length numSteps.
		"""
		res = []
		try:
			for i in range(numSteps):
				res.append(self.driveOn())
		except ProcessAbsorbed:
			pass
		return [a for a in interleave(
			map(self._getStateName, self.stateSeq), res)]


def _argmax(aSeq):
	return aSeq.index(max(aSeq))


class FancyHMM(HiddenMarkovModel):
	"""This class adds Viterbi- and Forward methods to a Hidden Markov
	Model.
	"""
	def _computeViterbiTrellis(self, observation):
		"""returns the phi and psi arrays for the viterbi algorithms.
		"""
		phi = [self.initDf.getDist()]
		psi = [[-1]*self.numStates]
		states = range(self.numStates)
		for ob in observation:
			curProbs, curPsis = [], []
			prevProbs = phi[-1]
			for newState in states:
				maxProb, srcState = 0, -1
				for oldState in states:
					curProb = (prevProbs[oldState]*self.tf[oldState][newState]*
						self.emissionMatrix[oldState][newState][ob])
					if curProb>maxProb:
						maxProb, srcState = curProb, oldState
				curProbs.append(maxProb)
				curPsis.append(srcState)
			phi.append(curProbs)
			psi.append(curPsis)
		return phi, psi


	def getMostLikelyPath(self, observation):
		"""runs a Viterbi algorithm for observation and returns the
		most likely path together with its total probability.
		"""
		observation = _outputToOutputNums(observation)
		phi, psi = self._computeViterbiTrellis(observation)
		lastState = _argmax(phi[-1])
		totalProb = phi[-1][lastState]
		stateSeq = [lastState]
		while psi:
			lastState = psi[-1][lastState]
			stateSeq.append(lastState)
			del psi[-1]
		del stateSeq[-1]
		stateSeq.reverse()
		stateSeq = map(self._getStateName, stateSeq)
		return totalProb, stateSeq


def _processOneRun(run, stateTransitions):
	"""analyses run (see estimateModelParameters) and enters the result
	into stateTransitions, which is a dictionary of source states
	with dictionaries as values.  The latter have the target states
	as keys and dictionaries of emissions and their counts as values.
	"""
	procState = 1
	lastSymbol = None
	for item in run:
		if procState:
			if lastSymbol is not None:
				targetDict = stateTransitions.setdefault(lastState, {}).setdefault(
					item, {})
				targetDict[lastSymbol] = targetDict.get(lastSymbol, 0)+1
			lastState = item
			procState = 0
		else:
			lastSymbol = item
			procState = 1


def _normalizeDictAndReturnTotal(dist):
	"""changes dist in place such that the sum of its values is one.
	It returns the total of all values prior to normalization.  What
	a mess of an interface.
	"""
	total = float(sum(dist.values()))
	for key in dist:
		dist[key] /= total
	return total


def estimateModelParameters(sampleRuns):
	"""ML-estimates transition probabilities and output probabilities
	from a list of sample runs of an HMM.  Since we're doing arc
	emission, these sample runs are simply sequences of state-emission-state
	etc.
	"""
	stateTransitions = {}
	for run in sampleRuns:
		_processOneRun(run, stateTransitions)
	emProbs = []
	transProbs = []
	for src, dests in stateTransitions.iteritems():
		transProbsForState = {}
		for dest, ems in dests.iteritems():
			transProbsForState[dest] = _normalizeDictAndReturnTotal(ems)
			emProbs.append("+%s-%s+ %s"%(src, dest, str(ems)))
		_normalizeDictAndReturnTotal(transProbsForState)
		transProbs.append("'%s': %s,"%(src, str(transProbsForState)))
	print "{\n%s\n}"%"\n".join(transProbs)
	print "---\n---"
	print "+-+\n".join(emProbs)


def initStuff():
	"""does option parsing and setup.  We return the non-option arguments 
	we don't use (i.e., all) and the options object.
	"""
	from optparse import OptionParser
	from time import time
	parser = OptionParser()
	parser.add_option("-s", "--random-seed", dest="randomSeed",
		help="set random number seed")
	options, args = parser.parse_args()
	if options.randomSeed is None:
		options.randomSeed = "%.2f"%time()
	random.seed(options.randomSeed)
	print "Random seed is", options.randomSeed
	return options, args


def probB(srcFName):
	hmm = HiddenMarkovModel(open(srcFName).read())
	hmm.setUpInitDf()
	print " ".join(hmm.getAnOutput(30))


def probC(srcFName):
	hmm = HiddenMarkovModel(open(srcFName).read())
	results = []
	for i in range(100):
		hmm.setUpInitDf()
		results.append(hmm.getAnOutput(30))
	estimateModelParameters(results)


def probD(srcFName):
	hmm = FancyHMM(open(srcFName).read())
	hmm.setUpInitDf()
	stateSeq, obs = deInterleave(hmm.getAnOutput(30))
	print " ".join(stateSeq)
	print " "+(" ".join(obs))
	print " ".join(hmm.getMostLikelyPath(obs)[1])


if __name__=="__main__":
	import sys
	opts, args = initStuff()
	probD(args[0])
